from datetime import datetime, timedelta

from django.db.models import Q

from .utils import RANGE_DATE_OPERATORS, PREVIOUS_RANGE_DATE_OPERATORS


class DatesMixin:
    # Notice, the previous keyword is related to value charts
    previous = False

    def handle_hours_ago(self, value, field, operator, now):
        value = now + timedelta(hours=-int(value))
        if operator.startswith('before'):
            lookup_field = f'{field}__lt'
        else:
            lookup_field = f'{field}__gt'

        return Q(**{lookup_field: value})

    def handle_past_future(self, field, operator, now):
        if operator == 'past':
            lookup_field = f'{field}__lte'
        else:
            lookup_field = f'{field}__gte'
        return Q(**{lookup_field: now})

    def handle_range_date_operator(self, operator, field, value, now):
        start_of_today = now.replace(hour=0, minute=0, second=0, microsecond=0)
        end_of_today = now.replace(hour=23, minute=59, second=59, microsecond=999999)

        _from = start_of_today
        _to = now

        range_date_operator = RANGE_DATE_OPERATORS[operator]
        if self.previous:
            range_date_operator = PREVIOUS_RANGE_DATE_OPERATORS[operator]

        if 'from' in range_date_operator:
            _from = range_date_operator['from'](start_of_today, value)
        if 'to' in range_date_operator:
            _to = range_date_operator['to'](end_of_today, value)

        return Q(**{f'{field}__range': [_from, _to]})

    def handle_date_operator(self, operator, field, value, tz):
        now = datetime.now(tz)
        if operator in ('past', 'future'):
            return self.handle_past_future(field, operator, now)
        elif operator.endswith('x_hours_ago'):
            return self.handle_hours_ago(value, field, operator, now)
        else:
            return self.handle_range_date_operator(operator, field, value, now)



import copy
import sys
from datetime import datetime
from unittest import mock

import pytest
import pytz
from django.test import TransactionTestCase
from django.urls import reverse

from django_forest.tests.fixtures.schema import test_schema
from django_forest.tests.resources.views.list.test_list_scope import mocked_scope
from django_forest.utils.schema import Schema
from django_forest.utils.schema.json_api_schema import JsonApiSchema


# reset forest config dir auto import
from django_forest.utils.scope import ScopeManager


@pytest.fixture()
def reset_config_dir_import():
    for key in list(sys.modules.keys()):
        if key.startswith('django_forest.tests.forest'):
            del sys.modules[key]


@pytest.mark.usefixtures('reset_config_dir_import')
class ResourceListSmartSegmentViewTests(TransactionTestCase):
    fixtures = ['article.json', 'publication.json',
                'session.json',
                'question.json', 'choice.json',
                'place.json', 'restaurant.json',
                'student.json',
                'serial.json']

    @pytest.fixture(autouse=True)
    def inject_fixtures(self, django_assert_num_queries):
        self._django_assert_num_queries = django_assert_num_queries

    def setUp(self):
        Schema.schema = copy.deepcopy(test_schema)
        Schema.add_smart_features()
        Schema.handle_json_api_schema()
        self.url = reverse('django_forest:resources:list', kwargs={'resource': 'tests_question'})
        self.client = self.client_class(
            HTTP_AUTHORIZATION='Bearer eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpZCI6IjUiLCJlbWFpbCI6Imd1aWxsYXVtZWNAZm9yZXN0YWRtaW4uY29tIiwiZmlyc3RfbmFtZSI6Ikd1aWxsYXVtZSIsImxhc3RfbmFtZSI6IkNpc2NvIiwidGVhbSI6Ik9wZXJhdGlvbnMiLCJyZW5kZXJpbmdfaWQiOjEsImV4cCI6MTYyNTY3OTYyNi44ODYwMTh9.mHjA05yvMr99gFMuFv0SnPDCeOd2ZyMSN868V7lsjnw')

    def tearDown(self):
        # reset _registry after each test
        JsonApiSchema._registry = {}
        ScopeManager.cache = {}

    @mock.patch('jose.jwt.decode', return_value={'id': 1, 'rendering_id': 1})
    @mock.patch('django_forest.utils.permissions.datetime')
    @mock.patch('django_forest.utils.scope.datetime')
    def test_get(self, mocked_scope_datetime, mocked_datetime, mocked_decode):
        mocked_scope_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 23, 582772, tzinfo=pytz.UTC)
        mocked_datetime.now.return_value = datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
        ScopeManager.cache = {
            '1': {
                'scopes': mocked_scope,
                'fetched_at': datetime(2021, 7, 8, 9, 20, 22, 582772, tzinfo=pytz.UTC)
            }
        }
        response = self.client.get(self.url, {
            'fields[tests_question]': 'id,topic,question_text,pub_date,foo,bar',
            'fields[topic]': 'name',
            'segment': 'Best questions',
            'page[number]': '1',
            'page[size]': '15',
            'timezone': 'Europe/Paris',
        })
        data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(data, {
            'data': [
                {
                    'type': 'tests_question',
                    'attributes': {
                        'pub_date': '2021-06-02T13:52:53.528000+00:00',
                        'question_text': 'what is your favorite color?',
                        'foo': 'what is your favorite color?+foo',
                        'bar': 'what is your favorite color?+bar'
                    },
                    'id': 1,
                    'links': {
                        'self': '/forest/tests_question/1'
                    },
                    'relationships': {
                        'topic': {
                            'data': None,
                            'links': {
                                'related': '/forest/tests_question/1/relationships/topic'
                            }
                        }
                    },
                },
            ]
        })



from datetime import timedelta
import math


def handle_previous_quarter_utils(x):
    current_quarter = math.ceil(x.month / 3)
    last_quarter = 4 if current_quarter == 1 else current_quarter - 1
    previous_last_quarter = 4 if last_quarter == 1 else last_quarter - 1
    year = x.year - 1 if current_quarter == 1 else x.year

    return current_quarter, last_quarter, previous_last_quarter, year


def get_current_month(current_quarter, last_quarter, previous):
    month = current_quarter
    if previous:
        month = last_quarter
    return month


def get_last_quarter_month(last_quarter, previous_last_quarter, previous):
    month = last_quarter
    if previous:
        month = previous_last_quarter
    return month


def handle_previous_quarter_from(x, current=False, previous=False):
    current_quarter, last_quarter, previous_last_quarter, year = handle_previous_quarter_utils(x)
    if current:
        month = get_current_month(current_quarter, last_quarter, previous)
    else:
        month = get_last_quarter_month(last_quarter, previous_last_quarter, previous)
    return x.replace(day=1, month=(month * 3) - 2, year=year)


def handle_previous_quarter_to(x, previous=False):
    _, last_quarter, previous_last_quarter, year = handle_previous_quarter_utils(x)
    month = last_quarter
    if previous:
        month = previous_last_quarter
    return x.replace(day=1, month=(month * 3) + 1, year=year) + timedelta(days=-1)


def get_yesterday(time_frame):
    return {
        'from': lambda x, y: x + timedelta(days=-(time_frame)),
        'to': lambda x, y: x + timedelta(days=-(time_frame))
    }


def get_previous_week(time_frame):
    return {
        'from': lambda x, y: x + timedelta(weeks=-(time_frame)) + timedelta(days=-x.weekday()),
        'to': lambda x, y: x + timedelta(weeks=-(time_frame)) + timedelta(days=6 - x.weekday())
    }


def get_previous_month(x):
    return (x.replace(day=1) + timedelta(days=-1)).replace(day=1)


def get_previous_year(x):
    return (x.replace(day=1, month=1) + timedelta(days=-1)).replace(day=1, month=1)


RANGE_DATE_OPERATORS = {
    'today': {
        'to': lambda x, y: x
    },
    'yesterday': get_yesterday(1),
    'previous_x_days': {
        'from': lambda x, y: x + timedelta(days=-int(y)),
        'to': lambda x: x + timedelta(days=-1)
    },
    'previous_x_days_to_date': {
        'from': lambda x, y: x + timedelta(days=-(int(y) - 1)),
    },
    'previous_week': get_previous_week(1),
    'previous_week_to_date': {
        'from': lambda x, y: x + timedelta(days=-x.weekday()),
    },
    'previous_month': {
        'from': lambda x, y: get_previous_month(x),
        'to': lambda x, y: x.replace(day=1) + timedelta(days=-1)
    },
    'previous_month_to_date': {
        'from': lambda x, y: x.replace(day=1),
    },
    'previous_quarter': {
        'from': lambda x, y: handle_previous_quarter_from(x),
        'to': lambda x, y: handle_previous_quarter_to(x)
    },
    'previous_quarter_to_date': {
        'from': lambda x, y: handle_previous_quarter_from(x, current=True),
    },
    'previous_year': {
        'from': lambda x, y: get_previous_year(x),
        'to': lambda x, y: x.replace(day=1, month=1) + timedelta(days=-1)
    },
    'previous_year_to_date': {
        'from': lambda x, y: x.replace(day=1, month=1)
    },
}

PREVIOUS_RANGE_DATE_OPERATORS = {
    'today': get_yesterday(1),
    'yesterday': get_yesterday(2),
    'previous_x_days': {
        'from': lambda x, y: x + timedelta(days=-int(y * 2)),
        'to': lambda x: x + timedelta(days=-1)
    },
    'previous_x_days_to_date': {
        'from': lambda x, y: x + timedelta(days=-((int(y) - 1) * 2)),
    },
    'previous_week': get_previous_week(2),
    'previous_week_to_date': {
        'from': lambda x, y: x + timedelta(weeks=-1) + timedelta(days=-x.weekday()),
    },
    'previous_month': {
        'from': lambda x, y: (get_previous_month(x) + timedelta(days=-1)).replace(day=1),
        'to': lambda x, y: get_previous_month(x) + timedelta(days=-1)
    },
    'previous_month_to_date': {
        'from': lambda x, y: get_previous_month(x),
    },
    'previous_quarter': {
        'from': lambda x, y: handle_previous_quarter_from(x, previous=True),
        'to': lambda x, y: handle_previous_quarter_to(x, previous=True)
    },
    'previous_quarter_to_date': {
        'from': lambda x, y: handle_previous_quarter_from(x, current=True, previous=True),
    },
    'previous_year': {
        'from': lambda x, y: get_previous_year(get_previous_year(x)),
        'to': lambda x, y: get_previous_year(x) + timedelta(days=-1)
    },
    'previous_year_to_date': {
        'from': lambda x, y: get_previous_year(x)
    },
}

DATE_OPERATORS = ['past', 'future', 'today', 'before_x_hours_ago', 'after_x_hours_ago'] \
    + list(RANGE_DATE_OPERATORS.keys())

PREVIOUS_DATE_OPERATOR = ['today', 'yesterday',
                          'previous_week', 'previous_month', 'previous_quarter', 'previous_year',
                          'previous_week_to_date', 'previous_month_to_date', 'previous_quarter_to_date',
                          'previous_year_to_date',
                          'previous_x_days', 'previous_x_days_to_date']
